# THANKS to https://github.com/AlexxIT/XiaomiGateway3 for the HASS implementation
# which I used as a base for the FHEM implementation

# current version is based on https://github.com/AlexxIT/XiaomiGateway3/tree/360c6d8ecc41f70ecf4f536e51bf379383f56200

import asyncio
import functools
import logging

from .. import fhem, utils
from .. import generic
from .core.gateway3 import GatewayEntry
from .core.utils import update_zigbee_firmware

DOMAINS = [
    "binary_sensor",
    "climate",
    "cover",
    "light",
    "remote",
    "sensor",
    "switch",
    "alarm_control_panel",
    "device_tracker",
]


class xiaomi_gateway3(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.gw = None
        self.devices = {}
        self.fhempy_devices = {}

        attr_conf = {"disable": {"options": "0,1", "default": "0", "format": "int"}}
        self.set_attr_config(attr_conf)

        set_conf = {
            "activate_zigbee2mqtt": {
                "help": (
                    "You need to install zigbee2mqtt aftewards<br>"
                    "=> https://www.zigbee2mqtt.io/getting_started/running_zigbee2mqtt.html#2-installing<br>"
                    "Set following lines in configuration.yaml<br>"
                    " serial:<br>"
                    "   port: tcp://GATEWAY_IP_ADDRESS:8888<br>"
                    "   adapter: ezsp<br>"
                    " mqtt:<br>"
                    "   client_id: zigbee_pi<br>"
                    "PLEASE CHECK THE LOG FILE AFTER CLICKING HERE, "
                    "DO NOT DISCONNECT GATEWAY FROM POWER"
                )
            },
            "deactivate_zigbee2mqtt": {},
        }
        self.set_set_config(set_conf)

    @property
    def gateway3(self):
        return self.gw.gateway3

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        self.hash = hash

        if self._attr_disable == 1:
            return

        if len(args) < 5:
            return "Usage: define devname fhempy xiaomi_gateway3 <IP> <TOKEN>"

        if await fhem.AttrVal(self.hash["NAME"], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon mqtt")
        await fhem.readingsSingleUpdateIfChanged(hash, "state", "disconnected", 1)

        logging.getLogger("fhempy.lib.xiaomi_gateway3.core.gateway3").setLevel(
            logging.ERROR
        )

        hash["HOST"] = args[3]
        hash["TOKEN"] = args[4]

        self.host = args[3]
        self.token = args[4]

        self.create_async_task(self.connect_gw())

        return ""

    async def set_activate_zigbee2mqtt(self, hash, params):
        self.create_async_task(update_zigbee_firmware(self.host, True))
        await fhem.readingsSingleUpdateIfChanged(hash, "zigbee2mqtt", "on", 1)

    async def set_deactivate_zigbee2mqtt(self, hash, params):
        self.create_async_task(update_zigbee_firmware(self.host, False))
        await fhem.readingsSingleUpdateIfChanged(hash, "zigbee2mqtt", "off", 1)

    async def set_attr_disable(self, hash):
        if self._attr_disable and self.gw is not None:
            await self.gw.stop()
            self.cancel_async_task(self.is_connected_task)
        else:
            self.create_async_task(self.connect_gw())

    async def register_device(self, fhempy_device, handler):
        did = fhempy_device.did
        self.fhempy_devices[did] = fhempy_device
        # wait till gateway is ready
        while self.gw is None:
            await asyncio.sleep(3)
        # if did == "lumi.0":
        #    self.gw.add_stats(did, handler)
        await asyncio.sleep(3)
        if did in self.devices and "init" in self.devices[did]:
            await fhempy_device.initialize(self.devices[did])

    async def connect_gw(self):
        await asyncio.sleep(0)
        self.gw = FhempyGateway(self.logger)
        z2m = await fhem.ReadingsVal(self.hash["NAME"], "zigbee2mqtt", "off")
        if z2m == "on":
            await self.gw.create_gateway(self.hash, self.host, self.token, zha=True)
        else:
            await self.gw.create_gateway(self.hash, self.host, self.token, zha=False)
            # prepare domains
            for domain in DOMAINS:
                self.gw.add_setup(domain, self.create_device)
        # start check task
        self.is_connected_task = self.create_async_task(self.is_connected())
        # run gateway
        self.gw.start()

    async def is_connected(self):
        last_was_connected = False
        while True:
            if await self.gw.is_connected():
                if last_was_connected is False:
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "connected", 1
                    )
                    last_was_connected = True
            else:
                if last_was_connected is True:
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "state", "disconnected", 1
                    )
                    last_was_connected = False
            await asyncio.sleep(60)

    async def create_device(self, gw, device, type):
        self.logger.debug(f"Check if device {device['did']} exists")
        did = device["did"]
        self.devices[did] = device
        if not await fhem.checkIfDeviceExists(
            self.hash, "PYTHONTYPE", "xiaomi_gateway3_device", "DID", did
        ):
            devname = (
                "".join(filter(str.isalnum, str(device["model"]))) + "_" + device["mac"]
            )
            await fhem.CommandDefine(
                self.hash,
                devname
                + " fhempy xiaomi_gateway3_device "
                + self.hash["NAME"]
                + " "
                + did,
            )
        else:
            if "init" in device and did in self.fhempy_devices:
                if did == "lumi." + device["mac"]:
                    device["version"] = self.gw.version()
                await self.fhempy_devices[did].initialize(device)
        # wait for fhem
        await asyncio.sleep(2)


class FhempyGateway:
    def __init__(self, logger):
        self.loop = asyncio.get_event_loop()

    async def create_gateway(self, hash, host, token, **options):
        self.gw = GatewayEntry(host=host, token=token, **options)

    @property
    def gateway3(self):
        return self.gw

    def add_setup(self, domain, handler):
        async def setup(gw, device, attr):
            await handler(gw, device, attr)

        self.gw.add_setup(domain, setup)

    def version(self):
        return self.gw.ver

    def start(self):
        self.gw.start()

    async def stop(self):
        await self.gw.stop()

    async def is_connected(self):
        return await self.gw.check_port(23)



import asyncio
import base64
import hashlib
import pathlib
from datetime import datetime

import aiohttp

from .. import fhem, generic


class github_backup(generic.FhemModule):

    headers = {
        "Authorization": "",
        "Accept": "application/vnd.github+json",
    }

    def __init__(self, logger):
        super().__init__(logger)

        self.gh_token_ready = asyncio.Event()

        attr_config = {
            "backup_time": {
                "default": "03:38",
                "format": "time",
                "help": "Daily backup at a specific time. Format HH:MM, default 03:38.",
            },
            "github_token": {"default": "", "help": "Personal github token"},
            "backup_files": {
                "default": (
                    "fhem.cfg,configDB.db,configDB.conf,log/fhem.save,"
                    + ".fhempy/zigbee2mqtt/data/configuration.yaml,"
                    + "esphome_config/*.yaml"
                ),
                "help": (
                    "Comma separated list of files to backup.<br>"
                    + "Default:<br>"
                    + "fhem.cfg,configDB.db,configDB.conf,log/fhem.save,"
                    + ".fhempy/zigbee2mqtt/data/configuration.yaml,"
                    + "esphome_config/*.yaml"
                ),
            },
        }
        self.set_attr_config(attr_config)

        set_config = {
            "backup_now": {},
        }
        self.set_set_config(set_config)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 5:
            return (
                "Usage: define my_backup fhempy github_backup"
                + " https://github.com/xxx/fhem_backup master_fhem_rpi"
            )

        self.url = args[3]
        self.gh_user = self.url.split("/")[-2]
        self.gh_repo = self.url.split("/")[-1]
        self.directory = args[4]

        icon = await fhem.AttrVal(self.hash["NAME"], "icon", "")
        if icon == "":
            await fhem.CommandAttr(self.hash, f"{self.hash['NAME']} icon system_backup")
            await fhem.CommandAttr(
                self.hash,
                f"{self.hash['NAME']} devStateIcon "
                + "failed:message_attention ok:message_ok",
            )

        # do this to check if gh token is set
        await self.set_attr_github_token(hash)

        self.create_async_task(self.update_static_readings())
        self.create_async_task(self.backup_loop())

    async def set_backup_now(self, hash, params):
        self.create_async_task(self.do_backup())

    async def set_attr_github_token(self, hash):
        if self._attr_github_token == "":
            await fhem.readingsSingleUpdate(
                hash, "state", "Please set github_token attribute", 1
            )
            return

        self.gh_token_ready.set()
        github_backup.headers["Authorization"] = f"token {self._attr_github_token}"

    async def github_get(self, url):
        ret = None
        async with aiohttp.ClientSession(trust_env=True) as session:
            async with session.get(url, headers=github_backup.headers) as resp:
                if resp.status < 400:
                    ret = await resp.json()
                else:
                    self.logger.error(
                        f"Failed to get {url} with HTTP error {resp.status}"
                    )
        return ret

    async def github_put(self, url, json_data):
        async with aiohttp.ClientSession(trust_env=True) as session:
            async with session.put(
                url, json=json_data, headers=github_backup.headers
            ) as resp:
                if resp.status < 400:
                    return True
                else:
                    self.logger.error(
                        f"Failed to put {url} with HTTP error {resp.status}"
                    )
        return False

    async def b64encode_file(self, filename):
        fh = open(filename, "rb")
        f_content = fh.read()
        fh.close()
        header = f"blob {len(f_content)}\0"
        gh_sha_content = header.encode("utf-8") + f_content
        sha = hashlib.sha1(gh_sha_content).hexdigest()
        b64_bytes = base64.b64encode(f_content)
        return b64_bytes.decode("ascii"), sha

    async def get_sha_from_file(self, filename):
        resp = await self.github_get(
            f"https://api.github.com/repos/{self.gh_user}/"
            + f"{self.gh_repo}/contents/{self.directory}/{filename}"
        )
        if resp is not None and "sha" in resp:
            return resp["sha"]
        return None

    async def upload_file(self, filename):
        try:
            # get current sha
            f_sha = await self.get_sha_from_file(filename)

            content, sha = await self.b64encode_file(filename)

            if f_sha == sha:
                # no update required
                return True

            # upload file
            data_msg = {
                "message": f"{self.directory} - {filename} backup",
                "content": content,
            }
            if f_sha:
                data_msg["sha"] = f_sha

            upload_resp = await self.github_put(
                f"https://api.github.com/repos/{self.gh_user}/"
                + f"{self.gh_repo}/contents/{self.directory}/{filename}",
                data_msg,
            )
            if upload_resp is False:
                raise Exception(f"Failed to upload file {filename}")

            return True
        except Exception:
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, f"{filename}_backup", "failed", 1
            )
            self.logger.exception(f"Failed to upload file {filename}")
        return False

    async def backup_loop(self):
        await self.gh_token_ready.wait()
        while True:
            await self.do_backup()
            now = datetime.today()
            seconds_till_time = (self._attr_backup_time - now).seconds
            await asyncio.sleep(seconds_till_time)

    async def do_backup(self):
        if self._attr_github_token == "":
            return

        state_val = "ok"
        for file in self._attr_backup_files.split(","):
            filepaths = pathlib.Path(".").glob(file)
            for filepath in filepaths:
                if not filepath.is_file():
                    continue

                if not await self.upload_file(filepath):
                    state_val = "failed"

        await fhem.readingsSingleUpdate(self.hash, "state", state_val, 1)

    async def update_static_readings(self):
        await fhem.readingsSingleUpdateIfChanged(
            self.hash,
            "repository",
            f"<html><a href='{self.url}' target='_blank'>"
            + "Open repository (new tab/window)</a></html>",
            1,
        )



"""
Based on iphonedetect from Home Assistant
From : https://community.home-assistant.io/t/iphone-device-tracker-on-linux/13698
Tracks iPhones by sending a udp message to port 5353.
An entry in the arp cache is then made and checked.
"""

import asyncio
import functools
import socket
import subprocess

from .. import fhem, generic, utils

HOME_STATES = {
    2: "REACHABLE",
    8: "DELAY",
    # 4: "STALE",
}

CONST_MESSAGE = b"Marco"
CONST_MESSAGE_PORT = 5353


class Host:
    """Host object with arp detection."""

    @staticmethod
    def find_with_ip():
        """Queries the network neighbours and lists found IP's"""
        state_filter = " nud " + " nud ".join(HOME_STATES.values()).lower()
        cmd = f"ip neigh show {state_filter}".split()
        neighbours = subprocess.run(cmd, shell=False, capture_output=True, text=True)
        neighbours_ip = [_.split()[0] for _ in neighbours.stdout.splitlines()]
        return neighbours_ip

    @staticmethod
    def find_with_arp():
        """Queries the arp table and lists found IP's"""
        cmd = "arp -na"
        neighbours = subprocess.run(cmd, shell=True, capture_output=True, text=True)
        neighbours_ip = [
            _.split()[1][1:-1]
            for _ in neighbours.stdout.splitlines()
            if _.count(":") == 5
        ]
        return neighbours_ip


class arp_presence(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)

        attr_config = {
            "interval": {
                "default": 60,
                "format": "int",
                "help": "Change interval in seconds, default is 60s.",
            }
        }
        self.set_attr_config(attr_config)

        set_config = {"update": {}}
        self.set_set_config(set_config)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 4:
            return "Usage: define presence_my_iphone fhempy arp_presence IP"
        self.dev_ip = args[3]
        await self.setup_scanner()
        self.create_async_task(self.update_loop())

    async def set_update(self, hash, params):
        await self.update_once()

    def ping_device(self):
        """Send UDP message to probe device."""
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.settimeout(1)
            s.sendto(CONST_MESSAGE, (self.dev_ip, CONST_MESSAGE_PORT))
        self.logger.debug(f"Probe sent to {self.dev_ip}")

    async def update_loop(self):
        while True:
            await self.update_once()
            await asyncio.sleep(self._attr_interval)

    async def setup_scanner(self):
        """Set up the Host objects and return the update function."""

        if (
            subprocess.run("which ip", shell=True, stdout=subprocess.DEVNULL).returncode
            == 0
        ):
            self.logger.debug("Using 'IP' to find tracked devices")
            self._use_cmd_ip = True
        elif (
            subprocess.run(
                "which arp", shell=True, stdout=subprocess.DEVNULL
            ).returncode
            == 0
        ):
            self.logger.warn("Using 'ARP' to find tracked devices")
            self._use_cmd_ip = False
        else:
            self.logger.fatal("Can't get neighbours from host OS!")
            await fhem.readingsSingleUpdate(
                self.hash, "state", "no command arp and ip", 1
            )

    async def update_once(self):
        """Update all the hosts on every interval time."""
        await utils.run_blocking(functools.partial(self.ping_device))

        if self._use_cmd_ip:
            device_ips = Host.find_with_ip()
        else:
            device_ips = Host.find_with_arp()

        if self.dev_ip in device_ips:
            await fhem.readingsSingleUpdate(self.hash, "presence", "present", 1)
            await fhem.readingsSingleUpdate(self.hash, "state", "present", 1)
        else:
            await fhem.readingsSingleUpdate(self.hash, "presence", "absent", 1)
            await fhem.readingsSingleUpdate(self.hash, "state", "absent", 1)



import asyncio
import datetime

import bluetooth_adapters
from bluetooth_auto_recovery import recover_adapter
from fhempy.lib.generic import FhemModule

from .. import fhem


class ble_reset(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._hours = 24
        self._resettask = None
        self._attr_list = {"reset_time": {"default": "04:00", "format": "str"}}
        self.set_attr_config(self._attr_list)
        set_list_conf = {
            "interval": {"args": ["hours"], "options": "1h,2h,4h,8h,12h,24h,manual"},
            "resetnow": {},
        }
        self.set_set_config(set_list_conf)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        self._reset_time = datetime.datetime.strptime(self._attr_reset_time, "%H:%M")

        hours = await fhem.ReadingsVal(hash["NAME"], "interval", "24h")
        if hours == "manual":
            self._hours = 0
        else:
            self._hours = int(hours[:-1])
            self._resettask = self.create_async_task(self.ble_reset())

        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "interval", hours)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "active")
        await fhem.readingsEndUpdate(hash, 1)

    async def ble_reset(self):
        while True:
            if self._hours > 0:
                now = datetime.datetime.now()
                first_reset = datetime.datetime(
                    now.year,
                    now.month,
                    now.day,
                    self._reset_time.hour,
                    self._reset_time.minute,
                )
                # calculate next reset time
                if (now - first_reset).total_seconds() > 0:
                    next_reset = now + datetime.timedelta(
                        seconds=(self._hours * 3600)
                        - ((now - first_reset).seconds % (self._hours * 3600))
                    )
                else:
                    next_reset = now + datetime.timedelta(
                        seconds=1 + ((first_reset - now).seconds % (self._hours * 3600))
                    )
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash,
                    "nextreset",
                    f"{next_reset.hour:02}:{next_reset.minute:02}",
                    1,
                )
                await asyncio.sleep((next_reset - now).seconds)
            # do reset now
            await self.ble_reset_once()

            now = datetime.datetime.now()
            await fhem.readingsSingleUpdate(
                self.hash, "lastreset", f"{now.hour:02}:{now.minute:02}", 1
            )
            if self._hours == 0:
                return

    async def ble_reset_once(self):
        await self.do_ble_reset()
        now = datetime.datetime.now()
        await fhem.readingsSingleUpdate(
            self.hash, "lastreset", f"{now.hour:02}:{now.minute:02}", 1
        )

    async def do_ble_reset(self):
        try:
            adapter_list = await bluetooth_adapters.get_bluetooth_adapters()
            adapters = bluetooth_adapters.get_adapters()
            await adapters.refresh()
            adapter_details = adapters.adapters

            for adapter in adapter_list:
                await recover_adapter(adapter[:-1], adapter_details[adapter]["address"])
        except Exception:
            self.logger.exception("Failed to reset bluetooth")

    async def set_interval(self, hash, params):
        if self._resettask:
            self._resettask.cancel()

        hours = params["hours"]
        if hours == "manual":
            self._hours = 0
            await fhem.readingsSingleUpdate(hash, "nextreset", "-", 1)
        else:
            self._hours = int(hours[:-1])
        await fhem.readingsSingleUpdate(hash, "interval", hours, 1)

        self._resettask = self.create_async_task(self.ble_reset())

    async def set_resetnow(self, hash, params):
        self.create_async_task(self.ble_reset_once())

    async def set_attr_reset_time(self, hash):
        if self._resettask:
            self._resettask.cancel()
        self._reset_time = datetime.datetime.strptime(self._attr_reset_time, "%H:%M")
        self._resettask = self.create_async_task(self.ble_reset())
